module threemaxdiag

  INTEGER, PARAMETER, PUBLIC         :: dp=SELECTED_REAL_KIND(8)
  COMPLEX, PARAMETER, PUBLIC         :: ZERO = cmplx(0.0_dp, 0.0_dp), ONE=cmplx(1.0_dp,0.0_dp)
  PUBLIC                             :: matrix_max_diag
  
  
contains

  !DIAGONALIZA UNA MATRIZ "A" REAL*8 Y SIMÉTRICA MEDIANTE LOS PAQUETES LAPACK
  !N - Dimensión de la matriz
  !A - Matriz a diagonalizar
  !W - Valores propios
  !Z - Vectores propios
  subroutine diag_dps (N,A,W,Z)
    real(kind=dp), dimension(:,:), intent(in)        :: A
    integer, intent(in)                              :: N
    real(kind=dp), dimension(:), intent(out)         :: W
    real(kind=dp), dimension(:,:), intent(out)       :: Z

    real(kind=dp), allocatable, dimension(:)         :: WORK
    integer                                          :: INFO

    if (size(A)/=N**2 .or. size(W)/=N .or. size(Z)/=N**2) then
       print*, "ERROR. Las dimensiones no coinciden"
    else

       allocate(WORK(3*N))

       do i = 1, N
          do j = i, N
             Z(i,j) = A(i,j)
          end do
       end do

       call dsyev('V', 'U', N, Z, N, W, WORK, 3*N, INFO)

       if (INFO /= 0) then
          do i = 1, N
             do j = 1, N
                print *, A(i,j)
             end do
          end do
          print "(/,A)", "STOP"
          Z = 12321
       end if


    end if

  end subroutine diag_dps


  !DIAGONALIZA MÁXIMAMENTE 3 MATRICES COMPLEJAS QUE NO CONMUTAN. 
  ! N          dimensión de las matrices
  ! A, B, C    matrices a diagonalizar simultáneamente de manera aproximada
  ! X          transformación unitaria que resulta del proceso
  subroutine matrix_max_diag(N, A, B, C, X)
    
    integer, intent(in)                              :: N
    complex(kind=dp), dimension(:,:), intent(inout)  :: A, B, C
    complex(kind=dp), dimension(:,:), intent(out)    :: X

    real(kind=dp), dimension(3,3)                    :: G, G_eigvectors
    real(kind=dp), dimension(3)                      :: G_eigvalues
    complex(kind=dp), dimension(3)                   :: Ha, Hb, Hc
    complex(kind=dp), dimension(:,:), allocatable    :: Rotation, auxA1, auxB1, auxC1
    real(kind=dp), dimension(3)                      :: r
    real(kind=dp)                                    :: sigma, pre_sigma
    complex(kind=dp)                                 :: rot_c, rot_s
    integer                                          :: maxval, ieig, tt

    allocate(Rotation(N,N))

    allocate(auxA1(N,N), auxB1(N,N), auxC1(N,N))

    !INICIALIZACIÓN DE LA TRANSFORMACIÓN UNITARIA A LA UNIDAD
    X = cmplx(0.0_dp,0.0_dp)
    do i = 1, N
       X(i,i) = cmplx(1.0_dp, 0.0_dp)
    end do

    !SUMA DE LOS MÓDULOS AL CUADRADO DE LAS COMPONENTES EXTRADIAGONALES DE TODAS LAS MATRICES 
    !A DIAGONALIZAR. ESTA ES LA CANTIDAD A MINIMIZAR.
    sigma = 0.0_dp
    do i = 1, N
       do j = 1, N
          if (i/=j) then
             sigma = sigma + abs(A(i,j))**2.0_dp + abs(B(i,j))**2.0_dp + abs(C(i,j))**2.0_dp
          end if
       end do
    end do

    !CICLO PRINCIPAL. SE SALE CUANDO sigma CONVERJA.
    do loopnumber = 1, 1000
       print *, "ILOOP = ", loopnumber
       
       do i = 1, N
          do j = i+1, N
             
             auxA1 = matmul(matmul(X, A), transpose(conjg(X)))
             auxB1 = matmul(matmul(X, B), transpose(conjg(X)))
             auxC1 = matmul(matmul(X, C), transpose(conjg(X)))

             Ha(1) = auxA1(i,i) - auxA1(j,j)
             Ha(2) = auxA1(i,j) + auxA1(j,i)
             Ha(3) = cmplx(0.0_dp, 1.0_dp) * (auxA1(j,i) - auxA1(i,j))

             Hb(1) = auxB1(i,i) - auxB1(j,j)
             Hb(2) = auxB1(i,j) + auxB1(j,i)
             Hb(3) = cmplx(0.0_dp, 1.0_dp) * (auxB1(j,i) - auxB1(i,j))

             Hc(1) = auxC1(i,i) - auxC1(j,j)
             Hc(2) = auxC1(i,j) + auxC1(j,i)
             Hc(3) = cmplx(0.0_dp, 1.0_dp) * (auxC1(j,i) - auxC1(i,j))

             !CONSTRUCCIÓN DE LA MATRIZ "G"
             G(1,1) = real(conjg(Ha(1))*Ha(1) + conjg(Hb(1))*Hb(1) + conjg(Hc(1))*Hc(1))
             G(1,2) = real(conjg(Ha(1))*Ha(2) + conjg(Hb(1))*Hb(2) + conjg(Hc(1))*Hc(2))
             G(1,3) = real(conjg(Ha(1))*Ha(3) + conjg(Hb(1))*Hb(3) + conjg(Hc(1))*Hc(3))
             G(2,2) = real(conjg(Ha(2))*Ha(2) + conjg(Hb(2))*Hb(2) + conjg(Hc(2))*Hc(2))
             G(2,3) = real(conjg(Ha(2))*Ha(3) + conjg(Hb(2))*Hb(3) + conjg(Hc(2))*Hc(3))
             G(3,3) = real(conjg(Ha(3))*Ha(3) + conjg(Hb(3))*Hb(3) + conjg(Hc(3))*Hc(3))

             G(2,1) = G(1,2)
             G(3,1) = G(1,3)
             G(3,2) = G(2,3)

             !BÚSQUEDA DE SUS AUTOVALORES Y AUTOVECTORES
             call diag_dps(3, G, G_eigvalues, G_eigvectors)

             if (G_eigvectors(1,1) == 12321) then
                print *, G_eigvalues(3)
                print *, norm2(r)
                print *, r
                stop "ERROR EN LA BÚSQUEDA DE AUTOVECTORES EN LA MATRIZ 'G'"
             end if

             !SE ELIGE EL AUTOVECTOR, (x,y,z), ASOCIADO AL AUTOVALOR MÁS GRANDE CON TAL 
             !SIGNO QUE CUMPLA: x>0
             if(abs(G_eigvalues(3)) > epsilon(0.0)) then
                r = G_eigvectors(:,3)
                if(r(1)<0.0_dp) then
                   r = -r
                end if
             else
                print *, "INFO: 0 valued eigvalue.", G_eigvalues
                print *, G
                print *, i, j, auxC1(i,i), auxC1(j,j), C(i,i), C(j,j), ONE, X(i,i)
                stop
             end if
             
             !MATRIX DE ROTACIÓN: R(nk,n'k',c,s)
             Rotation = cmplx(0.0_dp, 0.0_dp)
             do m = 1, N
                Rotation(m,m) = cmplx(1.0_dp, 0.0_dp)
             end do
                             
             rot_s = cmplx(r(2), -r(3)) / sqrt(2.0_dp * norm2(r) * (r(1) + norm2(r)))
             rot_c = sqrt((r(1) + norm2(r)) / (2.0_dp * norm2(r)))
             
             Rotation(i,i) = rot_c
             Rotation(j,j) = conjg(rot_c)
             Rotation(i,j) = conjg(rot_s)
             Rotation(j,i) = -rot_s

             !SE ACUMULA LA ROTACIÓN, R, EN LA MATRIZ DE CAMBIO DE BASE, X.
             X = matmul(Rotation, X)
             
          end do
       end do

       
       !DIFERENCIA ENTRE LA sigma ANTERIOR Y LA ACTUAL. SI LA DIFERENCIA ES MENOR A LA 
       !TOLERANCIA, epsilon(0.0)~10^-7, SE SALE DEL CICLO PRINCIPAL
       pre_sigma = sigma

       auxA1 = matmul(matmul(X, A),  transpose(conjg(X)))
       auxB1 = matmul(matmul(X, B),  transpose(conjg(X)))
       auxC1 = matmul(matmul(X, C),  transpose(conjg(X)))
       sigma = 0.0_dp
       do i = 1, N
          do j = 1, N
             if (i/=j) then
                sigma = sigma + abs(auxA1(i,j))**2.0_dp + abs(auxB1(i,j))**2.0_dp + abs(auxC1(i,j))**2.0_dp
             end if
          end do
       end do

    
       print *, "PRESIGMA, SIGMA:", pre_sigma, ",", sigma
       print *, ""


       if (sigma * pre_sigma < 0 .or. abs(sigma-pre_sigma) <= epsilon(0.0)) then
          print *, "Se ha llegado a la convergencia"
          exit
       end if

    end do !FINAL DEL CICLO PRINCIPAL



    !SE ESCRIBEN EN PANTALLA LAS MATRICES DIAGONALIZADAS SIMULTÁNEAMENTE DE MANERA APROXIMADA.
    do i = 1, N
       print "(12(F8.4, F8.4, ','))", (auxA1(i,j), j=1, N)
    end do

    print *, ""
    
    do i = 1, N
       print "(12(F8.4, F8.4, ','))", (auxB1(i,j), j=1, N)
    end do

    print *, ""

    do i = 1, N
       print "(12(F8.4, F8.4, ','))", (auxC1(i,j), j=1, N)
    end do

  end subroutine matrix_max_diag
  

end module threemaxdiag
