!AUTOR:    MIKEL ROUCO MARTÍN
!EMAIL:    mikel.mrm@gmail.com
!
!DESCRIPCIÓN: Busca las funciones de Wannier máximamente localizadas de un cristal planteado mediante 
!          la diagonalización simultánea aproximada de las componentes del operador posición proyectado
!          a un subespacio recíproco. 

program wannier
  use threemaxdiag          !MÓDULO DONDE SE ENCUENTRA EL MÉTODO DE DIAGONALIZACIÓN SIMULTÁNEA APROXIMADO

  REAL(kind=dp), PARAMETER                       :: H = 1.0_dp / 3.0_dp, PI = acos(-1.0_dp)

  INTEGER                                        :: NBND, N1, N2, N3, NKS, NR1, NR2, NR3
  INTEGER, DIMENSION(:), ALLOCATABLE             :: units_bw
  REAL(kind=dp)                                  :: ALAT
  REAL(kind=dp), DIMENSION(3)                    :: dim_cell
  REAL(kind=dp), DIMENSION(:,:), ALLOCATABLE     :: K

  REAL(kind=dp)                                  :: modulo
  CHARACTER(len=256)                             :: file_name, formato
  INTEGER                                        :: bloch_unit, ibnd

  COMPLEX(kind=dp), DIMENSION(:,:), ALLOCATABLE  :: R, E, Mx, My, Mz, U
  REAL(kind=dp), DIMENSION(:), ALLOCATABLE       :: W
  REAL(kind=dp)                                  :: suma, fase, real1, im1, x, y, z
  COMPLEX(kind=dp)                               :: sum_wf


  !LECTURA DE LA INFORMACIÓN ACERCA DE LA ESTRUCTURA CRISTALINA CONSIDERADA
  open(unit=23,file="data.buffer",action="read",status="old")
  rewind(unit=23)
  read(unit=23,fmt=*) NBND, ALAT, dim_cell(:)
  read(unit=23,fmt=*) N1, N2, N3          !NÚMERO DE CELDAS EN CADA DIRECCIÓN DEL CRISTAL
  read(unit=23,fmt=*) NKS                 !NÚMERO DE POSIBLES VALORES DE k. (NKS = N1*N2*N3)

  !DIMENSIONES DE LA CELDILLA UNIDAD
  NR1 = nint(ALAT*dim_cell(1)/H)
  NR2 = nint(ALAT*dim_cell(2)/H)
  NR3 = nint(ALAT*dim_cell(3)/H)

  allocate( K(NKS,3) )
  do ik = 1, NKS
     read(unit=23,fmt=*) K(ik,:)
  end do
  close(unit=23)
  
  !SE CREAN Y NORMALIZAN LAS ONDAS DE BLOCH A PARTIR DE LAS FUNCIONES PERIÓDICAS, u_nk, 
  !OBTENIDAS MEDIANTE QUANTUM ESPRESSO Y SE ESCRIBEN EN LA CARPETA "U/".
  bloch_unit = 23
  do ibnd = 1, NBND
     do ik = 1, NKS

        call wave_module(ik, modulo)

        write(unit=file_name,fmt="('U/U',I3.3,I2.2,'.dat')") ik, ibnd
        open(unit=bloch_unit,file=file_name,action="write",status="replace")
        call normalizer(bloch_unit,ibnd,ik,modulo)
        close(unit=bloch_unit)

        print *, ik, modulo, trim(file_name)

     end do

     !SE BORRAN LOS ARCHIVOS DONDE SE GUARDABAN LAS u_nk
     call system('rm -rf fort.*')


     !CONSTRUCCIÓN DE LAS MATRICES A DIAGONALIZAR SIMULTÁNEAMENTE DE MANERA APROXIMADA: <k1|{x,y,z}|k2>
     allocate( Mx(NKS,NKS), My(NKS,NKS), Mz(NKS,NKS), U(NKS,NKS) )
     do ik1 = 1, NKS
        do ik2 = 1, NKS
           call position_matrices(ik1, ik2, K, Mx(ik1,ik2), My(ik1,ik2), Mz(ik1,ik2))
        end do
     end do


     !SE ESCRIBEN LAS MATRICES EN ARCHIVOS DENTRO DE LA CARPETA "M/"
     write(unit=file_name,fmt="('M/Mx',I3.3,'.dat')") ibnd 
     open(unit=23,file=file_name,action="write",status="replace")
     do i = 1, NKS
        write(23,"(12(F12.6,'+',F12.6,'i;',tr3))") (Mx(i,j), j=1,NKS) 
     end do
     close(23)

     write(unit=file_name,fmt="('M/My',I3.3,'.dat')") ibnd 
     open(unit=23,file=file_name,action="write",status="replace")
     do i = 1, NKS
        write(23,"(12(F12.6,'+',F12.6,'i;',tr3))") (My(i,j), j=1,NKS) 
     end do
     close(23)

     write(unit=file_name,fmt="('M/Mz',I3.3,'.dat')") ibnd 
     open(unit=23,file=file_name,action="write",status="replace")
     do i = 1, NKS
        write(23,"(12(F12.6,'+',F12.6,'i;',tr3))") (Mz(i,j), j=1,NKS) 
     end do
     close(23)


     !ALGORITMO DE DIAGONALIZACIÓN SIMULTÁNEA DE CARDOSO-SOULOUMIAC PARA OOBTENER
     !LA TRANSFORMACIÓN UNITARIA, U.
     call matrix_max_diag(NKS, Mx, My, Mz, U)

     !SE APLICA LA TRANSFORMACIÓN UNITARIA A LAS MATRICES Y SE VUELVEN A ESCRIBIR
     !LOS RESULTADOS EN LA CARPETA "M/".
     Mx = matmul( matmul(U,Mx), conjg(transpose(U)) )
     My = matmul( matmul(U,My), conjg(transpose(U)) )
     Mz = matmul( matmul(U,Mz), conjg(transpose(U)) )

     write(unit=file_name,fmt="('M/MxW',I3.3,'.dat')") ibnd 
     open(unit=23,file=file_name,action="write",status="replace")
     do i = 1, NKS
        write(23,"(12(F12.6,'+',F12.6,'i;',tr3))") (Mx(i,j), j=1,NKS) 
     end do
     close(23)

     write(unit=file_name,fmt="('M/MyW',I3.3,'.dat')") ibnd 
     open(unit=23,file=file_name,action="write",status="replace")
     do i = 1, NKS
        write(23,"(12(F12.6,'+',F12.6,'i;',tr3))") (My(i,j), j=1,NKS) 
     end do
     close(23)

     write(unit=file_name,fmt="('M/MzW',I3.3,'.dat')") ibnd 
     open(unit=23,file=file_name,action="write",status="replace")
     do i = 1, NKS
        write(23,"(12(F12.6,'+',F12.6,'i;',tr3))") (Mz(i,j), j=1,NKS) 
     end do
     close(23)


     write(unit=file_name,fmt="('M/U',I3.3,'.dat')") ibnd 
     open(unit=23,file=file_name,action="write",status="replace")
     do i = 1, NKS
        write(23,"(12(F12.6,'+',F12.6,'i;',tr3))") (U(i,j), j=1,NKS) 
     end do
     close(23)

     write(unit=file_name,fmt="('M/mU',I3.3,'.dat')") ibnd 
     open(unit=23,file=file_name,action="write",status="replace")
     do i = 1, NKS
        write(23,"(12(F12.6,'+',F12.6,'i;',tr3))") (abs(U(i,j))**2.0_dp, j=1,NKS) 
     end do
     close(23)



     !UTILIZANDO U, SE CONSTRUYEN LAS FWOL (EN "WF/") A PARTIR DE LAS ONDAS DE BLOCH (EN "U/")
     allocate(units_bw(NKS))

     do ibw = 1, NKS
        write(unit=file_name,fmt="('U/U',I3.3,I2.2,'.dat')") ibw, ibnd
        units_bw(ibw) = 1645+ibw
        open(unit=units_bw(ibw),file=file_name,action="read",status="old")
        rewind(units_bw(ibw))
     end do
      
     do iwf = 1, NKS
        write(unit=file_name,fmt="('WF/WF',I3.3,I2.2,'.dat')") iwf, ibnd
        print *, trim(file_name)
        open(unit=23,file=file_name,action="write",status="replace")
     
        do Gx = 0, N1-1
        do Gy = 0, N2-1
        do Gz = 0, N3-1

           do rz = 0, NR3-1
           do ry = 0, NR2-1
           do rx = 0, NR1-1
              sum_wf = cmplx(0.0_dp, 0.0_dp)
              do ibw = 1, NKS
                 read(unit=units_bw(ibw),fmt=*) real1, im1
                 fase = (2.0_dp*PI/ALAT) * (K(ibw,1)*(real(rx,kind=dp)/3.0_dp + real(Gx,kind=dp)*dim_cell(1)*ALAT) + K(ibw,2)*(real(ry,kind=dp)/3.0_dp + real(Gy,kind=dp)*dim_cell(2)*ALAT) + K(ibw,3)*(real(rz,kind=dp)/3.0_dp + real(Gz,kind=dp)*dim_cell(3)*ALAT))
                 sum_wf = sum_wf + U(iwf,ibw)*cmplx(cos(fase), sin(fase))*cmplx(real1, im1)
              end do
              write(unit=23,fmt="(3(F15.4), 3F)") real(rx)*H + real(Gx)*dim_cell(1)*ALAT, real(ry)*H + real(Gy)*dim_cell(2)*ALAT, real(rz)*H + real(Gz)*dim_cell(3)*ALAT, sum_wf, abs(sum_wf)**2.0_dp
           end do
           end do
           end do

           do ibw = 1, NKS
              rewind(units_bw(ibw))
           end do

        end do
        end do
        end do

        close(23)
        
     end do


  end do
  



contains

  !DADO UN ARCHIVO CON LA FUNCIÓN u_nk, CALCULA SU MÓDULO
  subroutine wave_module (ik, resultado)
    INTEGER, INTENT(IN)        :: ik
    REAL(kind=dp), INTENT(OUT) :: resultado
    
    complex(kind=dp)           :: sum
    real(kind=dp)              :: real1, im1 
    integer                    :: unit, rx, ry, rz

    unit = ik*100 + IBND
    REWIND(unit)
    
    sum = cmplx(0.0_dp, 0.0_dp)

    CELL: do rz = 0, NR3-1          !R_z
       do ry = 0, NR2-1             !R_y
          do rx = 0, NR1-1          !R_x
             read (unit=unit,fmt=*) real1, im1

             sum = sum + H**3.0_dp*abs(cmplx(real1, im1))**2.0_dp
             
          end do !R_x
       end do !R_y
    end do CELL !R_z


    resultado = sqrt(sum * N1 * N2 * N3)   !PORQUE TENEMOS QUE EXPANDIRLO A TODO EL CRISTAL

  end subroutine wave_module




  !CONSTRUYE LAS ONDAS DE BLOCH NORMALIZADAS Y LAS ESCRIBE EN UN ARCHIVO
  subroutine normalizer (w_unit, ibnd, ik, modulo)
    INTEGER, INTENT(IN)       :: ik, w_unit, ibnd
    REAL(kind=dp), INTENT(IN) :: modulo
    
    integer              :: r_unit, rx, ry, rz
    real(kind=dp)        :: aux_real, aux_im

    r_unit = ik*100 + ibnd
    REWIND(w_unit)
    REWIND(r_unit)

    do rz = 0, NR3-1
    do ry = 0, NR2-1
    do rx = 0, NR1-1
       
       read  (unit=r_unit,fmt=*) aux_real, aux_im

       write (unit=w_unit,fmt="(2F16.8)") aux_real/modulo, aux_im/modulo

    end do
    end do
    end do
  end subroutine normalizer



  !CÁLCULO DE <k1|{x,y,z}|k2>
  subroutine position_matrices(ik1, ik2, K, resx, resy, resz)
    INTEGER, INTENT(IN)                           :: ik1, ik2
    REAL(kind=dp), DIMENSION(:,:), INTENT(IN)     :: K
    COMPLEX(kind=dp), INTENT(OUT)                 :: resx, resy, resz

    integer                                       :: rx, ry, rz
    complex                                       :: sum, sumx, sumy, sumz
    real                                          :: real1, im1, real2, im2, fase
    character(len=20)                             :: filename1, filename2


    write (unit=filename1,fmt="('U/U',I3.3,I2.2,'.dat')") ik1, IBND
    open(unit=23,file=filename1,action="read",status="old")
    rewind(23)

    if(ik1 /= ik2) then
       write (unit=filename2,fmt="('U/U',I3.3,I2.2,'.dat')") ik2, IBND
       open(unit=24,file=filename2,action="read",status="old")
       rewind(24)
    end if
    
    sum   = cmplx(0.0_dp, 0.0_dp)
    sumx  = cmplx(0.0_dp, 0.0_dp)
    sumy  = cmplx(0.0_dp, 0.0_dp)
    sumz  = cmplx(0.0_dp, 0.0_dp)

    !INTEGRAL SOBRE UNA ÚNICA CELDA
    CELL: do rz = 0, NR3-1          !R_z

       do ry = 0, NR2-1             !R_y

          do rx = 0, NR1-1          !R_x
             read (unit=23,fmt=*) real1, im1

             if(ik1 == ik2) then
                real2 = real1
                im2 = im1
             else
                read (unit=24,fmt=*) real2, im2
             end if
             
             fase = 2.0_dp * PI * ((K(ik2,1)-K(ik1,1))*real(rx, kind=dp) + (K(ik2,2)-K(ik1,2))*real(ry, kind=dp) + (K(ik2,3)-K(ik1,3))*real(rz, kind=dp)) / (3.0_dp*ALAT)
             
             sum   = sum   + H**3.0_dp*cmplx(cos(fase), sin(fase))*cmplx(real2, im2)*cmplx(real1, -im1)
             sumx  = sumx  + H**3.0_dp*cmplx(cos(fase), sin(fase))*cmplx(real2, im2)*cmplx(real1, -im1) * real(rx,kind=dp) / 3.0_dp
             sumy  = sumy  + H**3.0_dp*cmplx(cos(fase), sin(fase))*cmplx(real2, im2)*cmplx(real1, -im1) * real(ry,kind=dp) / 3.0_dp
             sumz  = sumz  + H**3.0_dp*cmplx(cos(fase), sin(fase))*cmplx(real2, im2)*cmplx(real1, -im1) * real(rz,kind=dp) / 3.0_dp


          end do !R_x

       end do !R_y

    end do CELL !R_z


    !SE EXTIENDE LA INTEGRAL ANTERIOR A TODO EL CRISTAL
    resx  = cmplx(0.0_dp, 0.0_dp)
    resy  = cmplx(0.0_dp, 0.0_dp)
    resz  = cmplx(0.0_dp, 0.0_dp)
    
    do rk = 0, N3-1
       do rj = 0, N2-1
          do ri = 0, N1-1
             fase = 2.0_dp * PI * (real(ri*dim_cell(1), kind=dp) * (K(ik2,1)-K(ik1,1)) + real(rj*dim_cell(2), kind=dp) * (K(ik2,2)-K(ik1,2)) + real(rk*dim_cell(3), kind=dp) * (K(ik2,3)-K(ik1,3)))

             resx  = resx  + cmplx(cos(fase), sin(fase)) * (sumx + real(ri*dim_cell(1), kind=dp)*ALAT*sum)
             resy  = resy  + cmplx(cos(fase), sin(fase)) * (sumy + real(rj*dim_cell(2), kind=dp)*ALAT*sum)
             resz  = resz  + cmplx(cos(fase), sin(fase)) * (sumz + real(rk*dim_cell(3), kind=dp)*ALAT*sum)

          end do 
       end do
    end do

       
    close(23)
    close(24)

  end subroutine position_matrices


end program wannier
